package com.perforce.hws.automation;

import com.perforce.p4java.PropertyDefs;

import static com.perforce.hws.automation.OsUtil.*;

import java.nio.file.Paths;
import java.util.Properties;

/**
 * There's a lot of path and string conventions used in this project that,
 * in theory, is just configuration. Instead of using hard-coded strings
 * everywhere, expose them as a property of this object.
 */
public class AutomationConfig {

    /**
     * Defines the version of p4d we want to use. This requires having the
     * right build directory installed.
     * <p>
     * The location is a format that replaces the platform variable with our
     * current platform name as needed.
     * <p>
     * It will be assumed that Windows always appends the .exe extension to the
     * location.
     */
    public enum P4DVersion {
        MAINLINE("builds/main/p4-bin/bin.{platform}/p4d"),
        R2014_1("depot/r14.1/p4-bin/bin.{platform}/p4d"),
        R2014_2("depot/r14.2/p4-bin/bin.{platform}/p4d"),
        R2015_1("depot/r15.1/p4-bin/bin.{platform}/p4d"),
        R2015_2("depot/r15.2/p4-bin/bin.{platform}/p4d"),
        P2016_1("depot/r16.1/p4-bin/bin.{platform}/p4d");

        private String location;

        P4DVersion(String location) {
            this.location = location;
        }

        public String getLocation() {
            return location;
        }
    }

    // Where we stick an integrated runtime environment
    private String installDir;

    // Where we save cached versions of things
    private String cacheDir;

    // All application logs and stdout stderr redirects should go here
    private String logDir;

    private String p4dRoot;

    private String p4dLogPath;

    // stderr/stdout file from p4d
    private String p4dOutputPath;

    private String p4dPidPath;

    // The name of the p4d executable file
    private String p4dBinary;

    // Setup p4d to run with TLS (self-signed cert).
    private boolean enableP4dTls;

    private boolean enableUnicodeP4d;

    private String p4dFingerprint;

    private P4DVersion p4DVersion;

    private String ftpHost;
    private String ftpVersionDir;
    private String ftpDownloadPath;

    private String p4dServerUri;
    private Properties p4javaProps;
    private String localP4Port;

    private String superUser;
    private String superPassword;

    // URL to the local Helix Web Services instance.
    private String localHws;

    // stderr/stoud file for HWS
    private String hwsOutputPath;

    private String hwsPidPath;

    private String hwsConfig;
    private String hwsP4dConfigDir;

    private String mockRaymondPidPath;
    private String mockRaymondStderrPath;
    private String mockRaymondStdoutPath;
    private String mockRaymondLaunchOutputPath;

    private boolean gitFusionEnabled;

    private boolean cloudEnabled;

    private boolean coverageEnabled;


    /**
     * Generates a new config with most of our system defaults
     */
    public AutomationConfig() {
        if (isWindows()) {
            installDir = "C:\\hwstemp";
            cacheDir = "C:\\hwscache";
        } else {
            installDir = "/tmp/hws";
            cacheDir = "/tmp/cache-hws";
        }

        logDir = Paths.get(installDir, "logs").toString();
        p4dRoot = Paths.get(installDir, "p4droot").toString();
        p4dLogPath = Paths.get(p4dRoot, "p4d.log").toString();
        p4dOutputPath = Paths.get(logDir, "p4d.out").toString();
        p4dPidPath = Paths.get(installDir, "p4d.pid").toString();

        if (isWindows()) {
            p4dBinary = "p4d.exe";
        } else {
            p4dBinary = "p4d";
        }

        p4DVersion = P4DVersion.MAINLINE;

        enableP4dTls = false;

        enableUnicodeP4d = true;

        ftpHost = "ftp.perforce.com";
        // I'm having issues with creating stream depots with r15.2, due to
        // a likely bug in P4Java (ticket is forthcoming)
//        ftpVersionDir = "r15.2";
        ftpVersionDir = "r15.1";
        ftpDownloadPath = String.format("perforce/%s/bin.%s%s", ftpVersionDir,
                ftpOsName(), ftpOsArch());

        localP4Port = "localhost:1666";
        p4dServerUri = "p4java://localhost:1666";
        p4javaProps = new Properties();
        // Never ever touch local user files if possible. The server will need
        // to do this as well.
        //
        // The commons "ServerCOnnectionPoolFacade" also uses these properties:
        // - sockSoTimeout (RpcPropertyDefs.RPC_SOCKET_SO_TIMEOUT_NICK)
        p4javaProps.setProperty("useAuthMemoryStore", "1");
        p4javaProps.setProperty(PropertyDefs.PROG_NAME_KEY, "HWS AUTOMATION");
        p4javaProps.setProperty(PropertyDefs.PROG_VERSION_KEY, "6.6.6");

        superUser = "super";
        superPassword = "superuser1A!";

        localHws = "http://localhost:9000";

        hwsConfig = Paths.get(installDir, "config", "hws.conf").toString();
        hwsP4dConfigDir = Paths.get(installDir, "config", "p4d").toString();
        hwsOutputPath = Paths.get(logDir, "hws.out").toString();
        hwsPidPath = Paths.get(installDir, "hws.pid").toString();

        mockRaymondPidPath = Paths.get(installDir, "mock_raymond.pid").toString();
        mockRaymondLaunchOutputPath = Paths.get(logDir, "mock_raymond.startup").toString();
        mockRaymondStderrPath = Paths.get(logDir, "mock_raymond.err").toString();
        mockRaymondStdoutPath = Paths.get(logDir, "mock_raymond.out").toString();

        cloudEnabled = false;
        gitFusionEnabled = false;

        coverageEnabled = true;
    }

    //------------------------------------------------------------------------
    // Here be getter/setter magic
    //------------------------------------------------------------------------

    public String getCacheDir() {
        return cacheDir;
    }

    public void setCacheDir(String cacheDir) {
        this.cacheDir = cacheDir;
    }

    public String getInstallDir() {
        return installDir;
    }

    public void setInstallDir(String installDir) {
        this.installDir = installDir;
    }

    public String getLogDir() {
        return logDir;
    }

    public void setLogDir(String logDir) {
        this.logDir = logDir;
    }

    public String getP4dRoot() {
        return p4dRoot;
    }

    public void setP4dRoot(String p4dRoot) {
        this.p4dRoot = p4dRoot;
    }

    public String getFtpDownloadPath() {
        return ftpDownloadPath;
    }

    public void setFtpDownloadPath(String ftpDownloadPath) {
        this.ftpDownloadPath = ftpDownloadPath;
    }

    public String getFtpHost() {
        return ftpHost;
    }

    public void setFtpHost(String ftpHost) {
        this.ftpHost = ftpHost;
    }

    public String getFtpVersionDir() {
        return ftpVersionDir;
    }

    public void setFtpVersionDir(String ftpVersionDir) {
        this.ftpVersionDir = ftpVersionDir;
    }

    public String getP4dBinary() {
        return p4dBinary;
    }

    public void setP4dBinary(String p4dBinary) {
        this.p4dBinary = p4dBinary;
    }

    public String getP4dLogPath() {
        return p4dLogPath;
    }

    public void setP4dLogPath(String p4dLogPath) {
        this.p4dLogPath = p4dLogPath;
    }

    public String getP4dOutputPath() {
        return p4dOutputPath;
    }

    public void setP4dOutputPath(String p4dOutputPath) {
        this.p4dOutputPath = p4dOutputPath;
    }

    public String getP4dPidPath() {
        return p4dPidPath;
    }

    public void setP4dPidPath(String p4dPidPath) {
        this.p4dPidPath = p4dPidPath;
    }

    public String getP4dServerUri() {
        return p4dServerUri;
    }

    public void setP4dServerUri(String p4dServerUri) {
        this.p4dServerUri = p4dServerUri;
    }

    public Properties getP4javaProps() {
        return p4javaProps;
    }

    public void setP4javaProps(Properties p4javaProps) {
        this.p4javaProps = p4javaProps;
    }

    public P4DVersion getP4DVersion() {
        return p4DVersion;
    }

    public void setP4DVersion(P4DVersion p4DVersion) {
        this.p4DVersion = p4DVersion;
    }

    public String getSuperPassword() {
        return superPassword;
    }

    public void setSuperPassword(String superPassword) {
        this.superPassword = superPassword;
    }

    public String getSuperUser() {
        return superUser;
    }

    public void setSuperUser(String superUser) {
        this.superUser = superUser;
    }

    public String getLocalHws() {
        return localHws;
    }

    public void setLocalHws(String localHws) {
        this.localHws = localHws;
    }

    public String getHwsConfig() {
        return hwsConfig;
    }

    public void setHwsConfig(String hwsConfig) {
        this.hwsConfig = hwsConfig;
    }

    public String getHwsP4dConfigDir() {
        return hwsP4dConfigDir;
    }

    public void setHwsP4dConfigDir(String hwsP4dConfigDir) {
        this.hwsP4dConfigDir = hwsP4dConfigDir;
    }

    public String getHwsOutputPath() {
        return hwsOutputPath;
    }

    public void setHwsOutputPath(String hwsOutputPath) {
        this.hwsOutputPath = hwsOutputPath;
    }

    public String getHwsPidPath() {
        return hwsPidPath;
    }

    public void setHwsPidPath(String hwsPidPath) {
        this.hwsPidPath = hwsPidPath;
    }

    public String getLocalP4Port() {
        return localP4Port;
    }

    public void setLocalP4Port(String localP4Port) {
        this.localP4Port = localP4Port;
    }

    public String getMockRaymondPidPath() {
        return mockRaymondPidPath;
    }

    public void setMockRaymondPidPath(String mockRaymondPidPath) {
        this.mockRaymondPidPath = mockRaymondPidPath;
    }

    public String getMockRaymondStderrPath() {
        return mockRaymondStderrPath;
    }

    public void setMockRaymondStderrPath(String mockRaymondStderrPath) {
        this.mockRaymondStderrPath = mockRaymondStderrPath;
    }

    public String getMockRaymondStdoutPath() {
        return mockRaymondStdoutPath;
    }

    public void setMockRaymondStdoutPath(String mockRaymondStdoutPath) {
        this.mockRaymondStdoutPath = mockRaymondStdoutPath;
    }

    public String getMockRaymondLaunchOutputPath() {
        return mockRaymondLaunchOutputPath;
    }

    public void setMockRaymondLaunchOutputPath(String mockRaymondLaunchOutputPath) {
        this.mockRaymondLaunchOutputPath = mockRaymondLaunchOutputPath;
    }

    public boolean isCloudEnabled() {
        return cloudEnabled;
    }

    public void setCloudEnabled(boolean cloudEnabled) {
        this.cloudEnabled = cloudEnabled;
    }

    public boolean isGitFusionEnabled() {
        return gitFusionEnabled;
    }

    public void setGitFusionEnabled(boolean gitFusionEnabled) {
        this.gitFusionEnabled = gitFusionEnabled;
    }

    public boolean isCoverageEnabled() {
        return coverageEnabled;
    }

    public void setCoverageEnabled(boolean coverageEnabled) {
        this.coverageEnabled = coverageEnabled;
    }

    public boolean isEnableP4dTls() {
        return enableP4dTls;
    }

    public void setEnableP4dTls(boolean enableP4dTls) {
        this.enableP4dTls = enableP4dTls;
        this.p4dServerUri = "p4javassl://localhost:1666";
    }

    public String getP4dFingerprint() {
        return p4dFingerprint;
    }

    public void setP4dFingerprint(String p4dFingerprint) {
        this.p4dFingerprint = p4dFingerprint;
    }

    public boolean isEnableUnicodeP4d() {
        return enableUnicodeP4d;
    }

    public void setEnableUnicodeP4d(boolean enableUnicodeP4d) {
        this.enableUnicodeP4d = enableUnicodeP4d;
    }
}
