package com.perforce.hws.automation;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Map;

import com.perforce.hws.util.FilesUtils;
import com.perforce.hws.util.ProcessHelpers;

import static java.lang.String.format;

/**
 * Generates an RPM from the unix binary archive.
 */
public class CreateRPM implements
        FilesUtils,
        ProcessHelpers,
        VersionFileMethods {

    /**
     * Example: ./java ...CreateRPM [path]
     *
     * - path: The unix binary archive (tar.gz) file.
     *
     * @param args
     */
    public static void main(String args[]) {
        try {
        	new CreateRPM().run(args[0], args[1]);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    CreateRPM() {
    }

    public void run(String sourceArchivePath, String rpmArchivePath) throws Exception {
        withTempDir(temp -> {
            Files.createDirectories(optPerforcePath(temp));
            unpackArchive(optPerforcePath(temp), sourceArchivePath);
            String specPath = createSpecFile(temp);
            String rpmPath = runRpmBuild(temp.toAbsolutePath().toString(), specPath);
            Files.delete(Paths.get(specPath));

            // Move this file to our current working area, tar it up to make it easy to match
            Path rpmSource = Paths.get(rpmPath);
            Path rpmTarget = rpmSource.getFileName();
            Files.move(rpmSource, rpmTarget);
            System.out.println("Moved " + rpmSource + " to " + rpmTarget.toAbsolutePath());

            String out = execOutOrFail("tar", "-c", "-z", "-v",
                    "-f", rpmArchivePath,
                    rpmTarget.toString());
            System.out.println(out);
        });
    }

    public void unpackArchive(Path dir, String archivePath) throws IOException, InterruptedException {
        execOrFail("tar", "-x", "-z", "-C", dir.toAbsolutePath().toString(),
                "-f", archivePath);
    }

    private Path optPerforcePath(Path root) {
        return Paths.get(root.toAbsolutePath().toString(), "opt", "perforce");
    }

    private String createSpecFile(Path temp) throws IOException {
        Path versionFile = Paths.get(optPerforcePath(temp).toString(), "helix-web-services", "Version");
        String content = new String(Files.readAllBytes(versionFile));
        Map<String, String> versionInfo = parseVersionFile(content);

        String productVersion = "2016.1.0";

        StringWriter stringWriter = new StringWriter();
        PrintWriter writer = new PrintWriter(stringWriter);
        writer.println("%define home /opt/perforce/helix-web-services");
        writer.println("Name: helix-web-services");
        writer.println(format("Version: %s", productVersion));
        writer.println("Release: " + versionInfo.get("PATCHLEVEL"));
        writer.println("Summary: Helix Web Services, a middleware HTTP server for Perforce Helix");
        writer.println("License: MIT");
        writer.println("Distribution: RPM-based Linux");
        writer.println("Group: Other");
        writer.println("Packager: support+packaging@perforce.com");
        writer.println("Requires: java >= 1.8");

        writer.print("\n%description\n");
        writer.println("Helix Web Services provides an HTTP-based protocol to broker to other services");
        writer.println("in the Perforce Helix ecosystem.");

        writer.print("\n%files\n");
        writer.println("%dir %{home}");
        writer.println("%dir %{home}/tmp");
        writer.println("%dir %{home}/log");
        Files.walk(optPerforcePath(temp))
                .filter(Files::isRegularFile)
                .forEach(p -> {
                    if (p.getFileName().toString().equals("helix-web-services.conf") ||
                            p.getFileName().toString().equals("log4j.xml")) {
                        writer.print("%config(noreplace) ");
                    }
                    writer.println("\"/" + temp.relativize(p).toString() + "\"");
                });

        writer.print("\n%pre\n");
        writer.println("if [ \"$1\" = \"1\" ]; then");
        writer.println("    echo 'Preparing to install Helix Web Services...'");
        writer.println("elif [ \"$1\" = \"2\" ]; then");
        writer.println("    echo 'Preparing to upgrade Helix Web Services...'");
        writer.println("fi");

        writer.print("\n%post\n");
        writer.println("exec 2>%{home}/log/error.log");
        writer.println("%{home}/sbin/setup-helix-ws");
        writer.println("if [ $? -ne 0 ]; then");
        writer.println("printf \"Failed to run %{home}/sbin/setup-helix-ws. See %{home}/log/error.log for details.\" ");
        writer.println("exit $?;");
        writer.println("fi");
        writer.println("");
        writer.println("service helix-ws start");
        writer.println("if [ $? -ne 0 ]; then");
        writer.println("printf \"Failed to start helix-ws. See %{home}/log/error.log for details.\" ");
        writer.println("exit $?;");
        writer.println("fi");
        writer.println("");
        writer.println("echo \"::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\"");
        writer.println("echo \":: Thank you for installing Helix Web Services!\"");
        writer.println("echo \":: \"");
        writer.println("echo \":: To define Helix P4D servers that your Helix Web Services can communicate with,\"");
        writer.println("echo \":: create a file with configuration for each in the following directory and \"");
        writer.println("echo \":: restart your HWS service\"");
        writer.println("echo \"::   %{home}/etc/p4d/\"");
        writer.println("echo \":: \"");
        writer.println("echo \":: To configure the Helix Web Services service, change the following\"");
        writer.println("echo \":: configuration file and restart your HWS service:\"");
        writer.println("echo \"::   %{home}/etc/helix-web-services.conf\"");
        writer.println("echo \":: \"");
        writer.println("echo \":: Example configuration file can be found at:\"");
        writer.println("echo \"::   %{home}/etc/p4d/example\"");
        writer.println("echo \":: \"");
        writer.println("echo \":: To restart your HWS Service use the following command:\"");
        writer.println("echo \"::   sudo service helix-ws restart\"");
        writer.println("echo \"::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\"");
        writer.println("\n%preun\n");
        writer.println("if [ \"$1\" = \"0\" ] ; then");
        writer.println("  %{home}/sbin/uninstall-helix-web-services --installer");
        writer.println("fi");
        writer.flush();

        Path specFile = Files.createTempFile("rpm", ".spec");

        Files.write(specFile, stringWriter.getBuffer().toString().getBytes());

        return specFile.toString();
    }

    private String runRpmBuild(String buildRoot, String specPath) throws IOException, InterruptedException {
        String output = execOutOrFail("rpmbuild", "--buildroot=" + buildRoot,"-bb", specPath);
        System.out.println("rpmbuild:");
        System.out.println(output);

        return Arrays.stream(output.split("\n"))
                .filter(s -> s.startsWith("Wrote: "))
                .map(s -> s.substring("Wrote: ".length()))
                .findFirst()
                .get();
    }
}
