package com.perforce.hws.automation;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perforce.hws.deployment.PathConventions;
import com.perforce.hws.util.FilesUtils;
import com.perforce.hws.util.ProcessHelpers;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;

/**
 * Unlike our typical cleanup tasks, this cleans up after a system installation.
 * Thus, you need to run it with elevated privileges.
 */
public class DeploymentCleanup implements ProcessHelpers, OsHelpers, FilesUtils, PathConventions {

    static Logger logger = LoggerFactory.getLogger(DeploymentCleanup.class);

    public static void main(String args[]) {
        try {
            new DeploymentCleanup().run();
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(666);
        }
    }

    public DeploymentCleanup() {
    }

    public void run() throws Exception {
        shutdownHWS();
        uninstall();
        cleanDirectories();
    }

    private void shutdownHWS() throws IOException, InterruptedException {
        if (isLinux()) {
            shutdownHWSViaServiceStop();
        } else if (isOSX()) {
            shutdownHWSViaLaunchctl();
        } else if (isWindows()) {
            if (isActiveViaSc()) {
                shutdownViaSc();
                deleteViaSc();
            }
        } else {
            throw new IllegalStateException("Shutdown not supported on OS");
        }
    }

    private void shutdownHWSViaServiceStop() {
        try {
            logger.info("shutting down via service stop");
            String out = execOutOrFail("service", "helix-ws", "stop");
            System.out.println(out);
        } catch (IOException | IllegalStateException e) {
            logger.debug("ignoring failure trying to stop service", e);
        } catch (InterruptedException e) {
            logger.warn("interrupted while trying to stop service", e);
        }
    }

    private void shutdownHWSViaLaunchctl() throws IOException, InterruptedException {
        execOutOrFail("launchctl", "unload",
                plistPath().toString());
    }

    private void shutdownViaSc() throws IOException, InterruptedException {
        String scPath = Paths.get(System.getenv("windir"), "system32", "sc.exe").toString();
        try {
            execOutOrFail(scPath, "stop", "helix-ws");
        } catch (IllegalStateException e) {
            logger.info("Ignoring failure trying to stop HWS via sc.exe", e);
        }
    }

    private void deleteViaSc() throws IOException, InterruptedException {
        String scPath = Paths.get(System.getenv("windir"), "system32", "sc.exe").toString();
        try {
            execOutOrFail(scPath, "delete", "helix-ws");
        } catch (IllegalStateException e) {
            logger.info("Ignoring failure trying to delete HWS via sc.exe", e);
        }
    }

    private boolean isActiveViaSc() throws IOException, InterruptedException {
        String scPath = Paths.get(System.getenv("windir"), "system32", "sc.exe").toString();
        String output;
        try {
            output = execOutOrFail(scPath, "query", "helix-ws");
//            logger.info("output of query: {}", output);
            return output.contains("SERVICE_NAME: helix-ws");
        } catch (IllegalStateException e) {
            return false;
        }
    }

    private void uninstall() throws IOException, InterruptedException {
        if (isLinux()) {
            if (isUbuntu()) {
                logger.info("looking for installation in ubuntu");
                if (isInstalledViaDpkg()) {
                    logger.info("removing helix-web-services ubuntu packages");
                    execOrFail("dpkg", "-r", "helix-web-services");
                    execOrFail("dpkg", "-P", "helix-web-services");
                }
            } else if (isCentOS()) {
                logger.info("looking for installation in centos");
                if (isInstalledViaRpm()) {
                    logger.info("removing helix-web-services CentOS package");
                    execOrFail("rpm", "-e", "helix-web-services");
                }
            } else {
                throw new IllegalStateException("Unknown Linux variant");
            }
        } else if (isOSX() || isWindows()) {
            logger.info("uninstallation unnecessary");
        } else {
            throw new IllegalStateException("Uninstall not supported on OS: "
                    + osName());
        }
    }

    private boolean isInstalledViaDpkg() throws IOException, InterruptedException {
        logger.debug("calling dpkg -l");
        // This seems to hang.
        String out = execOutOrFail("dpkg", "-l");
        return Arrays.stream(out.split("\n"))
                .map(line -> line.split("\\s+"))
                .filter(parts -> parts.length > 1)
                .map(parts -> parts[1])
                .anyMatch(s -> s.contains("helix-web-services"));
    }

    private boolean isInstalledViaRpm() {
        try {
            String installedPackages = execOutOrFail("rpm", "-qa");
            return Arrays.stream(installedPackages.split("\n"))
                    .anyMatch(s -> s.contains("helix-web-services"));
        } catch (IOException | IllegalStateException e) {
            return false;
        } catch (InterruptedException e) {
            throw new IllegalStateException(e);
        }
    }

    private void removeSystemUser(String username) throws IOException, InterruptedException {
        if (isLinux()) {
            removeSystemUserViaUserdel(username);
        } else if (isOSX()) {
            if (userExistsViaDscl(username)) {
                removeSystemUserViaDscl(username);
            }
        } else if (isWindows()) {
            logger.info("No system user is typically created on Windows, skipping");
        } else {
            throw new IllegalStateException("removeSystemUser does not work on OS: " + osName());
        }
    }

    private void removeSystemUserViaUserdel(String username) throws InterruptedException {
        logger.info("removing system user {} via userdel", username);
        try {
            execOrFail("userdel", username);
        } catch (IOException | IllegalStateException e) {
            logger.debug("ignoring userdel failure", e);
        }
    }

    private boolean userExistsViaDscl(String username) throws InterruptedException {
        try {
            execOutOrFail("dscl . -read /Users/" + username);
            return true;
        } catch (IOException | IllegalStateException e) {
            return false;
        }
    }

    private void removeSystemUserViaDscl(String username) throws InterruptedException {
        try {
            execOrFail("dscl", ".", "-delete", "/Groups/" + username);
            execOrFail("dscl", ".", "-delete", "/Users/" + username);
        } catch (IOException | IllegalStateException e) {
            logger.debug("ignoring dscl failures removing " + username, e);
        }
    }

    private void cleanDirectories() throws IOException {
        if (isLinux() || isOSX()) {
            logger.info("removing common Unix installation locations");
            cleanUnixDirectories();
        } else if (isWindows()) {
            cleanWindowsDirectories();
        } else {
            throw new IllegalStateException("cleanDirectories not supported on OS: " + osName());
        }
    }

    private void cleanUnixDirectories() throws IOException {
        checkRemove("/opt/perforce/helix-web-services");
        checkRemove("/var/log/helix-web-services");
        checkRemove("/etc/init.d/helix-ws");
    }

    private void cleanWindowsDirectories() throws IOException {
        checkRemove("C:\\helix-web-services");
    }

    private void checkRemove(String first, String... more) throws IOException {
        Path p = Paths.get(first, more);
        if (Files.exists(p, LinkOption.NOFOLLOW_LINKS)){
            if (Files.isDirectory(p)) {
                logger.info("deleting directory {}", p);
                rmtree(p.toAbsolutePath().toString());
            } else {
                logger.info("deleting file {}", p);
                Files.delete(p);
            }
        } else {
            logger.info("file {} does not exist", p);
        }
    }
}
