package com.perforce.hws.automation;

import java.io.IOException;
import java.util.Arrays;

import com.perforce.hws.util.ProcessHelpers;

public interface GitFusionManagement extends UsesAutomationConfig, OsHelpers, ProcessHelpers {

    default void setupGitFusion() throws IOException, InterruptedException {
        if (!isGitFusionInstalled()) {
            installGitFusion();
        }
    }

    default boolean isGitFusionInstalled() throws IOException, InterruptedException {
        if (!isLinux()) {
            return false;
        }
        if (isUbuntu()) {
            String out = execOutOrFail("dpkg", "-l");
            return Arrays.stream(out.split("\n"))
                        .map(line -> line.split("\\s+"))
                    .filter(p -> p.length > 1)
                    .map(p -> p[1])
                    .anyMatch(s -> s.contains(gitFusionPackageName()));
        } else {
            try {
                String allPackages = execOutOrFail("rpm", "-qa");
                return Arrays.stream(allPackages.split("\n"))
                        .anyMatch(s -> s.contains(gitFusionPackageName()));
            } catch (IOException | IllegalStateException e) {
                return false;
            } catch (InterruptedException e) {
                throw new IllegalStateException(e);
            }
        }
    }

    /**
     * Note that this assumes that the Perforce package repository has already been configured.
     *
     * @throws IOException
     */
    default void installGitFusion() throws IOException, InterruptedException {
        if (isUbuntu()) {
        	execOutOrFail("apt-get", "update");
            String out = execOutOrFail("apt-get", "install", "-y", "helix-git-fusion");
            logger.info("apt-get install helix-git-fusion: {}", out);
        } else {
            String out = execOutOrFail("yum", "install", "-y", "helix-git-fusion");
            logger.info("yum install helix-git-fusion: {}", out);
        }
        String configureOut =
                execOutOrFail("/opt/perforce/git-fusion/libexec/configure-git-fusion.sh",
                        "-n", // Non-interactive mode
                        "-m", // Monochrome
                        "--super", config().getSuperUser(),
                        "--superpassword", config().getSuperPassword(),
                        "--gfp4password", "gitfusion1A!",
                        "--server", "local",
                        "--p4port", "localhost:1666",
                        "--unknownuser", "pusher");
        logger.info("configure-git-fusion.sh: {}", configureOut);
    }

    default void removeGitFusion() throws IOException, InterruptedException {
        if (isUbuntu()) {
            String out = execOutOrFail("apt-get", "remove", "-y", "helix-git-fusion");
            logger.info("apt-get remove helix-git-fusion: {}", out);
            out = execOutOrFail("apt-get", "autoremove", "-y");
            logger.info("apt-get autoremove: {}", out);
            out = execOutOrFail("dpkg", "-P", "helix-git-fusion-3rdparty-libgit2");
            logger.info("dpkg -P helix-git-fusion-3rdparty-libgit2: {}", out);
        } else {
            String out = execOutOrFail("rpm", "-e", gitFusionPackageName());
            logger.info("rpm -e {}: {}", gitFusionPackageName(), out);
        }
    }

    default String gitFusionPackageName() {
        return "helix-git-fusion";
    }
}
