package com.perforce.hws.automation;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static com.perforce.hws.automation.OsUtil.nullPath;
import static com.perforce.hws.automation.ProcessUtil.isRunning;
import static com.perforce.hws.automation.ProcessUtil.pidOf;
import static com.perforce.hws.automation.ProcessUtil.waitForPidToQuit;

import java.io.File;
import java.io.IOException;
import java.net.Socket;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

/**
 * Methods that handle the background mock_raymond process.
 * <p>
 * This is a server application we use to test the integration with Helix Cloud
 * in HWS.
 */
public interface ManageMockRaymond extends
        UsesAutomationConfig, ProjectLocations {

    Logger logger = LoggerFactory.getLogger(ManageMockRaymond.class);

    default void startMockRaymond() throws IOException {
        if (isMockRaymondRunning()) {
            logger.info("mock_raymond already started");
            return;
        }

        if (Files.exists(Paths.get(config().getMockRaymondPidPath()))) {
            logger.debug("mock_raymond PID exists... deleting");
            Files.delete(Paths.get(config().getMockRaymondPidPath()));
        }

        String mockRaymondDir = locateMockRaymondDir();

        if (!Files.exists(Paths.get(mockRaymondDir))) {
            throw new IllegalStateException("mock_raymond does not exist at " +
                    mockRaymondDir);
        }

        ProcessBuilder pb = new ProcessBuilder(
                "bundle", "exec", "unicorn", "-c", "config/unicorn.rb", "-D");

        pb.directory(new File(mockRaymondDir));

        pb.environment().put("UNICORN_PID", config().getMockRaymondPidPath());
        pb.environment().put("RAILS_ENV", "test");
        pb.environment().put("MOCKRAYMOND_STDOUT_PATH", config().getMockRaymondStdoutPath());
        pb.environment().put("MOCKRAYMOND_STDERR_PATH", config().getMockRaymondStderrPath());
        pb.environment().put("P4PORT", config().getLocalP4Port());
        pb.environment().put("P4CHARSET", "auto");

        pb.redirectOutput(new File(config().getMockRaymondStdoutPath()));
        pb.redirectError(new File(config().getMockRaymondStderrPath()));
        pb.redirectInput(new File(nullPath()));

        Process process = pb.start();

        int pid = pidOf(process);

        if (pid != -1) {
            logger.info("Started mock_raymond at {}", pid);
            if (!process.isAlive()) {
                throw new IllegalStateException("mock_raymond failed to " +
                        "start with exit " + process.exitValue() + ": " +
                        pb.command());
            }
        } else {
            throw new IllegalStateException("Unable to launch or detect pid " +
                    "of mock_raymond: " + pb.command());
        }

        try {
            int retries = 60 * 10; // Wait no more than 1 minute
            while (0 < retries && !isMockRaymondRunning()) {
                Thread.sleep(100);
                retries--;
            }
        } catch (InterruptedException e) {
            throw new IllegalStateException(e);
        }
    }

    /**
     * Test to see if mock_raymond is responding at port 3000.
     *
     * @return true if server is responding to requests (immediately)
     */
    default boolean isMockRaymondRunning() {
        try (Socket s = new Socket("localhost", 3000)) {
            return true;
        } catch (IOException e) {
            //logger.debug("Ingoring socket error", e);
        }
        return false;
    }

    default void stopMockRaymond() throws IOException, InterruptedException {
        String pid = getMockRaymondPid();
        if (pid != null) {
            logger.debug("Stopping mock_raymond, sending SIGTERM to {}", pid);
            Runtime.getRuntime().exec("kill -SIGTERM " + pid);
            waitForPidToQuit(pid);
            if (isRunning(pid)) {
                throw new IllegalStateException("process " + pid + " is still running");
            }
        }
    }

    default String getMockRaymondPid() throws IOException {
        Path pidPath = Paths.get(config().getMockRaymondPidPath());
        if (Files.exists(pidPath)) {
            byte[] content = Files.readAllBytes(pidPath);
            return new String(content, "UTF-8");
        }
        return null;
    }

    /**
     * Detects where mock_raymond is within our current project directory,
     * mostly based on conventions within the project.
     *
     * @return The absolute path to the mock_raymond project directory.
     * @throws IOException
     */
    default String locateMockRaymondDir() throws IOException {
        String root = projectRootOrFail();
        return Paths.get(root, "mock_raymond")
                .toAbsolutePath()
                .toString();
    }
}
