package com.perforce.hws.automation;

import com.perforce.hws.p4base.CloseableServer;
import com.perforce.hws.p4base.HasGson;
import com.perforce.hws.p4base.P4Methods;
import com.perforce.hws.util.StreamReaderThread;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import static com.perforce.hws.automation.ProcessUtil.*;

/**
 * Methods related to managing a p4d instance.
 */
public interface ManageP4dMethods extends
        AutomationServerProvider,
        FilesystemMethods,
        HasGson,
        NetworkUtil,
        P4CommandLineMethods,
        P4Methods,
        ProjectLocations,
        UsesAutomationConfig {

    Logger logger = LoggerFactory.getLogger(ManageP4dMethods.class);

    default String p4dPath() throws IOException {
        String platform = OsUtil.ftpOsName() + OsUtil.ftpOsArch();
        String location = config().getP4DVersion().getLocation();
        location = location.replace("{platform}", platform);
        String path = Paths.get(clientRoot(), location).normalize().toAbsolutePath().toString();
        if (OsUtil.isWindows()) {
            path = path + ".exe";
        }
        return path;
    }

    /**
     * Creates a 'localhost' p4d configuration file.
     */
    default void addP4dConfig() {
        try {
            StringBuilder sb = new StringBuilder();
            sb.append("id: localhost\n");
            sb.append("name: localhost\n");
            sb.append("description: Local p4d server\n");
            sb.append("P4PORT: localhost:1666\n");
            sb.append("P4CHARSET: utf8\n");

            Path configDir = Paths.get(config().getHwsP4dConfigDir());

            if (!Files.exists(configDir)) {
                Files.createDirectories(configDir);
            }

            Path localConfigPath = Paths.get(configDir.toAbsolutePath().toString(), "localhost");
            Files.write(localConfigPath, sb.toString().getBytes());

            logger.info("Created server configuration for 'localhost' at {}",
                    localConfigPath.toAbsolutePath().toString());
        } catch (IOException io) {
            throw new UncheckedIOException(io);
        }
    }

    /**
     * Launches p4d that we've downloaded at localP4d() in installDir().
     * <p>
     * We detach stdout and stderr, because this can tie these processes to our
     * existing JVM runtime on Windows (surprise!).
     */
    default void startP4d() throws IOException, InterruptedException {
        if (!Files.exists(Paths.get(p4dPath()))) {
            throw new IOException("The p4d executable path does not " +
                    "exist: " + p4dPath() + ". Please check your client " +
                    "mappings and try again.");
        }

        if (config().isEnableP4dTls() && config().getP4dFingerprint() == null) {
            createPrivateKey();
        }

        boolean hasPidFileOption = true;
        if (config().getP4DVersion().equals(AutomationConfig.P4DVersion.R2014_1)) {
            hasPidFileOption = false;
        }

        List<String> command = new ArrayList<>();
        Collections.addAll(command,
                p4dPath(),
                "-r", config().getP4dRoot(),
                "-v", "server=3");
//                "-L", config().getP4dLogPath());

        command.add("-p");
        if (config().isEnableP4dTls()) {
            command.add("ssl::1666");
        } else {
            command.add("1666");
        }

        // We have problems with restarts on OS X.
        // We don't run single threaded mode by default, since that runs into problems when using
        // Git Fusion.
        if (isOSX()) {
            command.add("-f");
        }

        if (hasPidFileOption) {
            command.add("--pid-file=" + config().getP4dPidPath());
        }

        logger.info("Starting p4d with command {}", command.stream().collect(Collectors.joining(" ")));

        ProcessBuilder pb = new ProcessBuilder(command);

        if (config().isEnableP4dTls()) {
            pb.environment().put("P4SSLDIR", getSslkeysPath());
        }

        pb.redirectOutput(ProcessBuilder.Redirect.INHERIT);
        pb.redirectInput(new File(nullPath()));

        Process launched = pb.start();

        P4DOutputLogger stdoutLogger = new P4DOutputLogger(launched.getErrorStream());
        stdoutLogger.start();

        int pid = pidOf(launched);

        if (pid != -1) {
            logger.info("Started p4d at process {}", pid);
            if (!launched.isAlive()) {
                throw new IllegalStateException("p4d failed with exit " +
                        launched.exitValue() + ": " + pb.command());
            }

            waitForP4Info();

            if (!hasPidFileOption) {
                Files.write(Paths.get(config().getP4dPidPath()), Integer.toString(pid).getBytes());
            }

            String p4dPid = getP4dPid();
            logger.info("p4d pid {}", p4dPid);
            if (p4dPid == null) {
                throw new IllegalStateException("No pid found after startup");
            }
        } else {
            throw new IllegalStateException("Error launching or we could not " +
                    "detect the pid of the p4d command of: " + pb.command());
        }
    }

    class P4DOutputLogger extends Thread {

        private InputStream is;

        public P4DOutputLogger(InputStream is) {
            this.is = is;
        }

        @Override
        public void run() {
            try {
                InputStreamReader isr = new InputStreamReader(is);
                BufferedReader br = new BufferedReader(isr);
                String line = null;
                while ((line = br.readLine()) != null) {
                    logger.debug(line);
                }
            } catch (IOException e) {
                logger.error("Problem reading from p4d", e);
            }
            logger.info("stopping reading of p4d");
        }
    }

    default void waitForP4Info() {
        logger.info("Waiting for p4 info to return");
        int retries = 5 * 60; // 5 minutes
        while (0 < retries && !hasP4Info()) {
            retries--;
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            }
        }
    }

    default boolean hasP4Info() {
        try (CloseableServer server = obtain(null, null, null)) {
            return server.get().getServerInfo() != null;
        } catch (Exception e) {
            return false;
        }
    }

    /**
     * Stops the p4d indicated by the p4d at p4d.pid in the install directory.
     */
    default void stopP4d() throws IOException, InterruptedException {
        logger.info("stopping p4d");
        if (isWindows()) {
            String pid = getP4dPid();
            if (isRunning(pid)) {
                logger.info("Killing p4d running at {}", pid);
                Runtime.getRuntime().exec("taskkill /F /PID " + pid);
                waitForPidToQuit(pid);
                waitForPort1666();
                if (!isPortAvailable(1666)) {
                    throw new IllegalStateException("The port 1666 is not available after shutdown yet");
                }
            }
        } else {
            String pid = getP4dPid();
            if (pid != null) {
                if (isRunning(pid)) {
                    logger.info("Stopping p4d, sending SIGTERM to {}", pid);
                    Runtime.getRuntime().exec("kill -SIGTERM " + pid);
                    waitForPidToQuit(pid);
                    waitForPort1666();
                    if (isRunning(pid)) {
                        logger.info("Stopping p4d, sending SIGKILL to {}", pid);
                        Runtime.getRuntime().exec("kill -SIGKILL " + pid);
                        waitForPidToQuit(pid);
                    }
                    waitForPort1666();
                    if (!isPortAvailable(1666)) {
                        throw new IllegalStateException("The port 1666 is not available after shutdown yet");
                    }
                } else {
                    logger.info("p4d pid file exists but process {} is not running", pid);
                }
            }
        }

    }

    default void waitForPort1666() throws InterruptedException {
        logger.info("Waiting for 1666 (TCP) to become available locally");
        int retries = 5 * 60; // 5 minutes
        while (0 < retries && !isPortAvailable(1666)) {
            retries--;
            Thread.sleep(1000);
        }
    }

    default void stopViaP4AdminStop() throws IOException {
        // P4Java does not implement "p4 admin" commands.
        // We map in the p4 client and attempt to use it against our test server.
        // The annoying issue is that we need to execute it in a temporary location.
        withTempDir(tempDir -> {
            initializeP4Config(tempDir);
            if (loginAsSuper(tempDir)) {
                execCommandAsSuper(tempDir, p4Path(), "admin", "stop");
            } else {
                logger.warn("unable to login to issue 'p4 admin stop', will attempt to kill");
            }
        });
    }

    default String getP4dPid() throws IOException {
        Path pidPath = Paths.get(config().getP4dPidPath());
        if (Files.exists(pidPath)) {
            byte[] content = Files.readAllBytes(pidPath);
            return new String(content, "UTF-8").trim();
        }
        return null;
    }

    /**
     * Executes p4d with the "-xi" flag to upgrade the DB to unicode mode.
     *
     * @return
     * @throws IOException
     */
    default void switchToUnicode() throws IOException, InterruptedException {
        logger.info("Upgrading p4d to unicode mode");
        ProcessBuilder pb = new ProcessBuilder(p4dPath(),
                "-r", config().getP4dRoot(),
                "-v", "server=1",
                "-xi");

        pb.redirectErrorStream(true);
        pb.redirectInput(new File(nullPath()));

        Process launched = pb.start();

        int exitCode = launched.waitFor();

        if (exitCode != 0) {
            throw new IOException("unicode upgrade failed, see log at " +
                    config().getP4dOutputPath());
        }
    }

    default void resetToUnicode() throws IOException, InterruptedException {
        stopP4d();
        switchToUnicode();
        startP4d();
    }

    default void createPrivateKey() throws IOException, InterruptedException {
        Path sslkeys = Paths.get(getSslkeysPath());
        if (!Files.exists(sslkeys)) {
            logger.info("Creating P4SSLDIR {}", sslkeys);
            Files.createDirectories(sslkeys);
            if (!isWindows()) {
                setOwnerOnly(sslkeys);
            }
        }

        ProcessBuilder pb = new ProcessBuilder(p4dPath(), "-Gc");

        pb.environment().put("P4SSLDIR", getSslkeysPath());

        pb.redirectErrorStream(true);

        Process launched = pb.start();

        StreamReaderThread stdout = new StreamReaderThread(launched.getInputStream());
        stdout.start();

        int exit = launched.waitFor();

        stdout.join();

        if (exit != 0) {
            throw new IllegalStateException("p4d -Gc failed: " + stdout.getStringBuffer());
        }

        pb = new ProcessBuilder(p4dPath(), "-Gf");

        pb.environment().put("P4SSLDIR", getSslkeysPath());

        launched = pb.start();

        P4DOutputLogger stderrLogger = new P4DOutputLogger(launched.getErrorStream());
        stdout = new StreamReaderThread(launched.getInputStream());
        stderrLogger.start();
        stdout.start();

        exit = launched.waitFor();

        stderrLogger.join();
        stdout.join();

        if (exit != 0) {
            throw new IllegalStateException("p4d -Gf failed");
        }

        String output = stdout.getStringBuffer().toString();

        Matcher matcher = Pattern.compile("Fingerprint: (?<fingerprint>.*)\n").matcher(output);

        if (matcher.matches()) {
            config().setP4dFingerprint(matcher.group("fingerprint"));
            logger.info("Using fingerprint {}", config().getP4dFingerprint());
        } else {
            throw new IllegalStateException("Did not find new p4d fingerprint");
        }
    }

    default String getSslkeysPath() {
        return Paths.get(config().getP4dRoot(), "sslkeys").toAbsolutePath().toString();
    }
}
