package com.perforce.hws.automation;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perforce.hws.util.ProcessHelpers;
import com.perforce.hws.util.StreamUtils;

import java.io.IOException;
import java.io.PrintStream;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.stream.Collectors;

/**
 * P4Java can't do everything we need, in particular, any "admin" method isn't supported.
 * For automation at the moment, we drive everything via the command line application.
 */
public interface P4CommandLineMethods extends
        ProcessHelpers,
        ProjectLocations,
        StreamUtils,
        UsesAutomationConfig {

    Logger logger = LoggerFactory.getLogger(P4CommandLineMethods.class);

    default String p4Path() throws IOException {
        // This is freaking out using builds on EC machines.
//        String platform = OsUtil.ftpOsName() + OsUtil.ftpOsArch();
//        String location = config().getP4DVersion().getLocation();
//        location = location.replace("{platform}", platform);
//        location = location.replace("p4d", "p4");
//        String path = Paths.get(clientRoot(), location).normalize().toAbsolutePath().toString();
        String path = "p4";
        if (OsUtil.isWindows()) {
            path = path + ".exe";
        }
        return path;
    }

    default void initializeP4Config(Path tmpDir) {
        try {
            // Create .p4config
            StringBuilder sb = new StringBuilder();
            sb.append("P4USER=super\n");
            sb.append("P4PORT=" + config().getLocalP4Port() + "\n");
            sb.append("P4TICKETS=" + tmpDir.toAbsolutePath().toString() + "/.p4tickets\n");
            sb.append("P4ENVIRO=" + tmpDir.toAbsolutePath().toString() + "/.p4enviro\n");
            sb.append("P4CHARSET=auto\n");

            Path p4configPath = Paths.get(tmpDir.toAbsolutePath().toString(), ".p4config");

            Files.write(p4configPath, sb.toString().getBytes());
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    default boolean loginAsSuper(Path tmpDir)  {
        try {
            ProcessBuilder pb = new ProcessBuilder();
            pb.directory(tmpDir.toFile());
            pb.environment().clear();
            pb.environment().put("P4CONFIG", ".p4config");

            pb.command(p4Path(), "login");

            Process process = pb.start();

            StringBuilder sb = new StringBuilder();

            int ch;
            while ((ch = process.getInputStream().read()) != 0) {
                sb.append((char)ch);
                if (sb.toString().equals("Enter password:")) {
                    break;
                }
            }

            process.getOutputStream().write(config().getSuperPassword().getBytes());
            process.getOutputStream().write('\n');
            process.getOutputStream().flush();
            process.getOutputStream().close();

            int retVal = process.waitFor();

            String err = new String(readAll(process.getErrorStream()));
            String out = new String(readAll(process.getInputStream()));

            logger.debug("stderr\n{}", err);
            logger.debug("stdout\n{}", out);

            if (retVal != 0) {
                logger.error("login failed in " + tmpDir.toAbsolutePath().toString());
                System.exit(666);
                return false;
            }
            return true;
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    default void execCommandAsSuper(Path tmpDir, String... command) {
        try {
            ProcessBuilder pb = new ProcessBuilder();
            pb.directory(tmpDir.toFile());
            pb.environment().clear();
            pb.environment().put("P4CONFIG", ".p4config");

            pb.command(command);

            Process process = pb.start();

            int retVal = process.waitFor();

            String err = new String(readAll(process.getErrorStream()));
            String out = new String(readAll(process.getInputStream()));

            logger.debug("stderr\n{}", err);
            logger.debug("stdout\n{}", out);

            if (retVal != 0) {
                String cmd = Arrays.stream(command).collect(Collectors.joining(" "));
                throw new IllegalStateException("command " + cmd + " failed: " + retVal);
            }
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }
}
