package com.perforce.hws.automation;

import com.google.gson.Gson;
import com.perforce.hws.util.StreamReaderThread;
import com.sun.jna.Pointer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static com.perforce.hws.automation.OsUtil.isWindows;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Field;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Java has some "fun" limitations, like the fact you can't obviously get the
 * pid handle for a spawned process.
 */
public class ProcessUtil {

    static final Logger logger = LoggerFactory.getLogger(ProcessUtil.class);

    public static boolean isRunning(String pid) throws IOException, InterruptedException {
        if (isWindows()) {
            return isRunningViaTasklist(pid);
        } else {
            return isRunningViaPs(pid);
        }
    }

    private static boolean isRunningViaTasklist(String pid) throws IOException, InterruptedException {
        if (pid == null) {
            return false;
        }

        String tasklist = Paths.get(System.getenv("windir"), "system32", "tasklist.exe").toString();

        Process ps = Runtime.getRuntime().exec(tasklist + " /NH");

        StreamReaderThread stdout = new StreamReaderThread(ps.getInputStream());
        StreamReaderThread stderr = new StreamReaderThread(ps.getErrorStream());

        stdout.start();
        stderr.start();

        int exit = ps.waitFor();

        stdout.join();
        stderr.join();

        if (exit != 0) {
            throw new IllegalStateException("tasklist.exe /NH failed:\n" + stdout.getStringBuffer() +
                    "\nerr:\n" + stderr.getStringBuffer());
        }

        List<List<String>> columnList =
                Arrays.stream(stdout.getStringBuffer().toString().split("\n"))
                        .filter(l -> !l.trim().isEmpty())
                    .map(l -> Arrays.asList(l.split("\\s+")))
                    .collect(Collectors.toList());

        return columnList.stream().anyMatch(c -> c.size() > 1 && c.get(1).contains(pid));
    }

    private static boolean isRunningViaPs(String pid) throws IOException, InterruptedException {
        if (pid == null) {
            return false;
        }

        Process ps = Runtime.getRuntime().exec("ps ax -o pid");

        StreamReaderThread stdout = new StreamReaderThread(ps.getInputStream());
        StreamReaderThread stderr = new StreamReaderThread(ps.getErrorStream());

        stdout.start();
        stderr.start();

        int exit = ps.waitFor();

        stdout.join();
        stderr.join();

        if (exit != 0) {
            throw new IllegalStateException("ps aux failed:\n" + stdout.getStringBuffer() +
                    "\nerr:\n" + stderr.getStringBuffer() );
        }

        List<String> lines = Arrays.stream(stdout.getStringBuffer().toString().split("\n"))
                .map(s -> s.trim())
                .filter(s -> {
                    // Fun little way to check if something is an integer.
                    Scanner sc = new Scanner(s);
                    if (!sc.hasNextInt()) return false;
                    sc.nextInt();
                    return !sc.hasNext();
                })
                .collect(Collectors.toList());

        return lines.stream().anyMatch(s -> pid.equals(s));
    }

    public static void waitForPidToQuit(String pid) throws IOException {
        logger.debug("waiting for pid to quit: {}", pid);
        try {
            int retries = 5 * 60; // 5 minutes
            while (0 < retries && isRunning(pid)) {
                retries--;
                Thread.sleep(1000);
            }
        } catch (InterruptedException e) {
            throw new IllegalStateException(e);
        }
    }

    public static int pidOf(Process p) {
        if (isWindows()) {
            return pidOfWindows(p);
        } else {
            return pidOfUnix(p);
        }
    }

    public static int pidOfWindows(Process process) {
        // See http://www.golesny.de/p/code/javagetpid
        // DO NOT SHIP THIS CODE.
        int pid = -1;
        if (process.getClass().getName().equals("java.lang.Win32Process") ||
                process.getClass().getName().equals("java.lang.ProcessImpl")) {
            try {
                Field f = process.getClass().getDeclaredField("handle");
                f.setAccessible(true);
                long handl = f.getLong(process);

                Kernel32 kernel = Kernel32.INSTANCE;
                W32API.HANDLE handle = new W32API.HANDLE();
                handle.setPointer(Pointer.createConstant(handl));
                pid = kernel.GetProcessId(handle);
            } catch (Throwable e) {
            }
        }
        return pid;
    }

    public static int pidOfUnix(Process process) {
        // See http://www.golesny.de/p/code/javagetpid
        int pid = -1;
        if (process.getClass().getName().equals("java.lang.UNIXProcess")) {
            try {
                Field f = process.getClass().getDeclaredField("pid");
                f.setAccessible(true);
                pid = f.getInt(process);
            } catch (Throwable e) {
            }
        }
        return pid;
    }
}
